import anime from 'animejs'
import { ImageOptions } from './image'
import { Prize } from './prize'
import * as svg from './svg'
import { THEME_DARK, THEME_DEFAULT, THEME_LIGHT } from './theme'
import { degreeToRadians, polarToCartesian } from './utils'

interface WheelOptions {
  pos: [number, number]
  radius: number
  fontSize: number
  buttonText: string
  buttonWidth: number
  buttonFontSize: number
  textRotate: boolean
  limit: number
  duration: number
  turn: number
  draw: boolean
  clockwise: boolean
  theme: string
  image: ImageOptions
}

const BASE_BUTTON_DEGREE = 80
const TEXT_BOTTOM_PERCENTAGE = 0.6
const THEME_MAP = {
  default: THEME_DEFAULT,
  Default: THEME_DEFAULT,
  DEFAULT: THEME_DEFAULT,
  light: THEME_LIGHT,
  Light: THEME_LIGHT,
  LIGHT: THEME_LIGHT,
  dark: THEME_DARK,
  Dark: THEME_DARK,
  DARK: THEME_DARK,
}

const DEFAULT_OPTIONS: WheelOptions = {
  pos: [0, 0],
  radius: 100,
  fontSize: 16,
  buttonText: 'Draw',
  buttonWidth: 50,
  buttonFontSize: 0,
  textRotate: false,
  limit: 0,
  duration: 5000,
  turn: 4,
  clockwise: true,
  draw: true,
  theme: 'default',
  image: null,
}

class Wheel {
  private center = [0, 0]
  private isRunning = false
  private options: WheelOptions = DEFAULT_OPTIONS
  private repeatTimes = 0
  private rotation = 0
  private weight = []
  private weightSum = 0

  constructor(
    private el: HTMLElement,
    private data: Array<string | Prize>,
    options: Partial<WheelOptions> = DEFAULT_OPTIONS
  ) {
    const len = this.data.length
    if (len < 3 || len > 12) {
      throw 'data length must between 3 and 12'
    }

    this.prepareOptions(options)
    this.checkPrize()

    if (this.options.draw) this.draw()
  }

  private animeFunc() {
    const opt = this.option

    this._turntable.style['transform-origin'] = 'center'

    this._button.style.cursor = 'pointer'
    this._button.style['transform-origin'] = 'center'
    this._button.addEventListener('mouseover', () => {
      if (opt.onButtonHover && typeof opt.onButtonHover === 'function') {
        opt.onButtonHover(anime, this._button)
        return
      }
      anime({
        targets: this._button,
        scale: 1.2,
        duration: 500,
      })
    })
    this._button.addEventListener('mouseout', () => {
      anime({
        targets: this._button,
        scale: 1,
        duration: 500,
      })
    })
    this._button.addEventListener('click', () => {
      this._run()
    })
  }

  private checkPrize() {
    for (const i in this.data) {
      const data = this.data[i]
      if (typeof data === 'string') {
        this.data[i] = {
          text: data,
          chance: 1,
        }
      } else {
        this.data[i] = {
          ...data[i],
          text: data.text || i,
          chance: data.chance && data.chance > 0 ? data.chance : 1,
        }
      }

      this.weight.push(Number(this.data[i].chance))
      this.weightSum += Number(this.data[i].chance)
    }
  }

  private drawDefault() {
    if (this._turntable && this._button) return

    const opt = this.option
    const theme = themes[opt.theme] || themes.default
    opt.color = {
      ...theme,
      ...opt.color,
    }
    if (!opt.inRadius) {
      opt.inRadius = getInnerRadius(this.radius)
    } else if (opt.inRadius > this.radius) {
      opt.inRadius = this.radius
    }

    if (!this._turntable) this._drawTurntable()
    if (!this._button) this._drawButton()
  }

  private drawResource() {
    const opt = this.option
    const res = opt.image
    if (typeof res === 'object' && Object.keys(res).length > 0) {
      if (res.turntable && typeof res.turntable === 'string') {
        this._turntable = svg.image(res.turntable, {
          width: this.radius * 2,
          height: this.radius * 2,
          x: opt.pos[0],
          y: opt.pos[1],
        })
        this._svg.appendChild(this._turntable)
      }
      if (res.button && typeof res.button === 'string') {
        if (!res.offset || typeof res.offset !== 'number') res.offset = 0
        const size = [50, 50]
        const buttonHeight = (size[1] * opt.buttonWidth) / size[0]
        this._button = svg.image(res.button, {
          width: opt.buttonWidth,
          height: buttonHeight,
          x: this.center[0] - opt.buttonWidth / 2,
          y: this.center[1] + res.offset - buttonHeight / 2,
        })
        this._svg.appendChild(this._button)
      }
    }
  }

  private drawTurntable() {
    if (this._turntable) return

    const opt = this.option
    this._svg.appendChild(
      svg.circle(this.center[0], this.center[1], this.radius, {
        fill: opt.color.border,
      })
    )
    this._svg.appendChild(
      svg.circle(this.center[0], this.center[1], opt.inRadius, {
        fill: opt.color.prize,
      })
    )

    this._turntable = svg.g()
    this._svg.appendChild(this._turntable)
    for (const d of opt.data) {
      const [pathD, dLen] = describeArc(
        this.center[0],
        this.center[1],
        opt.inRadius,
        -this._deg / 2,
        this._deg / 2
      )

      const pie = svg.path(pathD, {
        fill: d.color || opt.color.prize,
        stroke: opt.color.line,
        strokeWidth: 2,
      })

      let fontSize = d.fontSize || opt.fontSize
      if (!fontSize) {
        let textSum = 0
        for (let i = 0; i < d.text.length; ++i) {
          if (d.text[i].match(/\w/)) {
            textSum += 1
          } else textSum += 2
        }
        fontSize =
          (BASE_FONT_SIZE * textSum) / 2 > dLen * opt.textBottomPercentage
            ? ((dLen * opt.textBottomPercentage) / textSum) * 2
            : BASE_FONT_SIZE
      }
      const text = svg.text(d.text, {
        x: this.center[0],
        y: opt.pos[1] + this.radius - opt.inRadius * opt.textBottomPercentage,
        fontSize,
        fill: d.fontColor || opt.color.prizeFont,
      })
      if (this.option.textRotate) {
        svg.rotate(
          text,
          degreeToRadians(90),
          this.center[0],
          opt.pos[1] + this.radius - opt.inRadius * opt.textBottomPercentage
        )
      }
      const gtext = svg.g([text])
      const g = svg.g([pie, gtext])
      svg.rotate(
        g,
        degreeToRadians(this._deg * opt.data.indexOf(d)),
        this.center[0],
        this.center[1]
      )
      this._turntable.appendChild(g)
      if (this.option.textRotate) {
        svg.translate(gtext, 0, -text.getComputedTextLength() / 2)
      } else {
        svg.translate(gtext, -text.getComputedTextLength() / 2, 0)
      }
    }
  }

  private drawButton() {
    if (this._button) return

    const opt = this.option
    if (opt.button && typeof opt.button === 'string') return

    const r = opt.buttonWidth / 2
    const center = this.center
    const deg = (180 - BASE_BUTTON_DEGREE) / 2
    const [pathArc, , , end] = describeArc(
      center[0],
      center[1],
      r,
      deg - 360,
      360 - deg
    )
    const top = [center[0], center[1] - r / Math.cos(degreeToRadians(deg))]
    const pathD = `${pathArc} L ${top[0]} ${top[1]} L ${end.x} ${end.y} L ${center[0]} ${center[1]}`
    const button = svg.path(pathD, { fill: opt.color.button })
    svg.dropShadow(this._svg, button, 0, 3, 3, 0.5)

    let text = null
    if (opt.buttonText !== '') {
      const maxLen = r * 2 * 0.8
      let fontSize = opt.buttonFontSize
      if (!fontSize) {
        let textSum = 0
        for (let i = 0; i < opt.buttonText.length; ++i) {
          if (opt.buttonText[i].match(/\w/)) {
            textSum += 1
          } else textSum += 2
        }
        fontSize =
          (BASE_FONT_SIZE * textSum) / 2 > maxLen
            ? (maxLen / textSum) * 2
            : BASE_FONT_SIZE
      }
      text = svg.text(opt.buttonText, {
        x: center[0],
        y: center[1],
        fontSize,
        fill: opt.color.buttonFont,
      })
    }

    this._button = svg.g([button, text])
    this._svg.appendChild(this._button)
    svg.translate(text, -text.getComputedTextLength() / 2, 0)
  }

  private prepareOptions(options: Partial<WheelOptions>) {
    this.options = {
      ...this.options,
      ...options,
    }

    this.center = this.options.pos.map((p) => p + this.options.radius)
  }

  private run() {
    if (this.isRunning) return

    const opt = this.option
    if (opt.limit > 0 && this.repeatTimes >= opt.limit) {
      opt.onFail && typeof opt.onFail === 'function' && opt.onFail()
      return
    }

    const runAnime = (pie) => {
      if (this.rotation > 0) {
        const revision = 360 - (this.rotation % 360)
        this.rotation += revision
      }
      this.rotation += getRotation(pie, this._deg, opt.turn)
      anime({
        targets: this._turntable,
        rotate: opt.clockwise
          ? this.rotation + 'deg'
          : '-' + this.rotation + 'deg',
        duration: opt.duration,
        begin: () => {
          this.isRunning = true
        },
        complete: () => {
          this.isRunning = false
          ++this.repeatTimes
          if (opt.onSuccess && typeof opt.onSuccess === 'function') {
            const d = opt.clockwise
              ? opt.data[(opt.data.length - pie) % opt.data.length]
              : opt.data[pie]
            opt.onSuccess(d)
          }
        },
      })
    }

    const random = Math.random() * this.weightSum
    let randomWeight = 0
    let pie = 0
    for (const i in this.weight) {
      randomWeight += this.weight[i]
      if (randomWeight > random) {
        pie = i
        runAnime(pie)
        break
      }
    }
  }

  draw() {
    const svgAttrs = {
      width: this.radius * 2,
      height: this.radius * 2,
    }
    if (this.el.tagName !== 'svg') {
      this._svg = svg.svg(svgAttrs)
      this.el.appendChild(this._svg)
    } else {
      this._svg = opt.el
      svg.useSVG(opt.el, svgAttrs)
    }

    this._deg = 360 / opt.data.length

    if (this.image) this._drawResource()
    this._drawDefault()

    this._animeFunc()
  }
}

export { Wheel }
